"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var debug = Debugger('akita:json-stream');
function uint8ArrayToString(data) {
    if (typeof data === 'string')
        return data;
    if (typeof Buffer !== 'undefined' && Buffer.isBuffer(data))
        return data.toString('utf-8');
    if (typeof TextDecoder !== 'undefined')
        return new TextDecoder('utf-8').decode(data);
    var str = '';
    for (var i = 0; i < data.length; i += 1) {
        str += String.fromCharCode(data[i]);
    }
    return str;
}
var JsonStream = (function () {
    function JsonStream(stream, reducer) {
        var _this = this;
        this.closed = false;
        this._stream = stream;
        this._queue = [];
        this._cache = '';
        this._reducer = reducer;
        this._listeners = {};
        var parseLine = function () {
            if (_this.closed)
                return;
            var index = _this._cache.indexOf('\n');
            if (index < 0)
                return;
            var line = _this._cache.substr(0, index).trim();
            _this._cache = _this._cache.substr(index + 1);
            if (line) {
                var json = void 0;
                try {
                    json = JSON.parse(line);
                }
                catch (e) {
                    if (_this.listenerCount('error')) {
                        _this.emit('error', e);
                    }
                    if (_this._reject) {
                        _this._reject(e);
                        _this._resolve = null;
                        _this._reject = null;
                        return;
                    }
                }
                if (_this._reducer && json.object) {
                    json.object = _this._reducer(json.object);
                }
                if (_this.listenerCount('data')) {
                    while (_this._queue.length) {
                        _this.emit('data', _this._queue.shift());
                    }
                    _this.emit('data', json);
                }
                if (_this._resolve) {
                    _this._resolve(json);
                    _this._resolve = null;
                    _this._reject = null;
                }
                else if (!_this.listenerCount('data')) {
                    _this._queue.push(json);
                }
            }
            parseLine();
        };
        this._handler = function (data) {
            if (_this.closed)
                return;
            var string = uint8ArrayToString(data);
            debug('receive', string);
            _this._cache += string;
            parseLine();
        };
        this._close = function () {
            if (_this.closed)
                return;
            _this.closed = true;
            if (_this.listenerCount('close')) {
                _this.emit('close');
            }
            if (_this._resolve) {
                _this._resolve(undefined);
                _this._resolve = null;
                _this._reject = null;
            }
            delete _this._stream;
            delete _this._reader;
            delete _this._cache;
            delete _this._reducer;
            delete _this._handler;
            _this._listeners = {};
        };
        if (stream.getReader) {
            this._reader = stream.getReader();
            var read_1 = function (_a) {
                var done = _a.done, value = _a.value;
                if (_this.closed)
                    return;
                _this._handler(value || (done && _this._cache ? '\n' : ''));
                if (done) {
                    _this._close();
                }
                else {
                    _this._reader.read().then(read_1);
                }
            };
            this._reader.read().then(read_1);
        }
        else {
            stream.on('data', this._handler);
            stream.on('close', this._close);
        }
    }
    JsonStream.prototype.on = function (name, fn) {
        if (!this._listeners[name]) {
            this._listeners[name] = [];
        }
        this._listeners[name].push(fn);
        return this;
    };
    JsonStream.prototype.emit = function (name) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        if (!this._listeners[name])
            return;
        this._listeners[name].forEach(function (fn) {
            fn.apply(null, args);
        });
    };
    JsonStream.prototype.listenerCount = function (name) {
        return (this._listeners[name] || []).length;
    };
    JsonStream.prototype.removeListener = function (name, fn) {
        if (!this._listeners[name])
            return;
        this._listeners[name] = this._listeners[name].filter(function (f) { return f !== fn; });
        return this;
    };
    JsonStream.prototype.removeAllListeners = function (name) {
        delete this._listeners[name];
        return this;
    };
    JsonStream.prototype.read = function () {
        var _this = this;
        if (this._queue.length) {
            return Promise.resolve(this._queue.shift());
        }
        if (this.closed)
            return Promise.reject(new Error('Can not read from closed stream.'));
        return new Promise(function (resolve, reject) {
            _this._resolve = resolve;
            _this._reject = reject;
        });
    };
    JsonStream.prototype.cancel = function () {
        if (this.closed)
            throw new Error('Can not cancel closed stream.');
        debug('cancel watch');
        if (this._stream.cancel) {
            this._stream.cancel();
        }
        else {
            this._stream.destroy();
            this._stream.removeListener('data', this._handler);
        }
        this._close();
    };
    return JsonStream;
}());
exports.default = JsonStream;
