"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var debug = Debugger('akita:query');
var Query = (function () {
    function Query(model, op) {
        this.model = model;
        this._filters = null;
        this._data = null;
        this._page = 0;
        this._limit = 0;
        this._sort = '';
        this._id = null;
        this._args = null;
        this._search = '';
        this._op = op;
        this._promise = null;
        this._lastField = '';
    }
    Query.prototype.arg = function (args, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._args) {
            this._args = {};
        }
        if (typeof args === 'object') {
            Object.assign(this._args, args);
        }
        else {
            this._args[args] = value;
        }
        return this;
    };
    Query.prototype.where = function (conditions, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._filters) {
            this._filters = {};
        }
        if (typeof conditions === 'object' && value === undefined) {
            this._filters = Object.assign(this._filters, conditions);
        }
        else if (typeof conditions === 'string') {
            if (value === undefined) {
                this._lastField = conditions;
            }
            else {
                this._filters[conditions] = value;
            }
        }
        else {
            throw new Error('Akita Error: invalid params for Query#where()');
        }
        return this;
    };
    Query.prototype.search = function (keyword) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._search = keyword;
        return this;
    };
    Query.prototype.__filter = function (type, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._filters) {
            this._filters = {};
        }
        if (!this._lastField) {
            throw new Error("Akita Error: you should invoke .where(string) before ." + type + "()");
        }
        if (type === 'eq') {
            this._filters[this._lastField] = value;
        }
        else {
            if (typeof this._filters[this._lastField] !== 'object') {
                this._filters[this._lastField] = {};
            }
            this._filters[this._lastField]["$" + type] = value;
        }
        return this;
    };
    Query.prototype.eq = function (value) {
        return this.__filter('eq', value);
    };
    Query.prototype.ne = function (value) {
        return this.__filter('ne', value);
    };
    Query.prototype.regex = function (value) {
        return this.__filter('regex', value);
    };
    Query.prototype.in = function (value) {
        return this.__filter('in', value);
    };
    Query.prototype.nin = function (value) {
        return this.__filter('nin', value);
    };
    Query.prototype.lt = function (value) {
        return this.__filter('lt', value);
    };
    Query.prototype.lte = function (value) {
        return this.__filter('lte', value);
    };
    Query.prototype.gt = function (value) {
        return this.__filter('gt', value);
    };
    Query.prototype.gte = function (value) {
        return this.__filter('gte', value);
    };
    Query.prototype.limit = function (size) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._limit = size;
        return this;
    };
    Query.prototype.page = function (value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._page = value;
        return this;
    };
    Query.prototype.sort = function (value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._sort = value;
        return this;
    };
    Query.prototype.exec = function () {
        var _this = this;
        if (this._result) {
            return this._result;
        }
        this._debug();
        var init = this._createInit();
        var path = init.path;
        delete init.path;
        var reducer;
        var M = this.model;
        var createRecord = function (data) { return new M(data, _this._params); };
        switch (this._op) {
            case 'remove':
                reducer = function (json) { return (json === null || json === void 0 ? void 0 : json.removed) || 0; };
                break;
            case 'findOne':
                reducer = function (json) {
                    if (!json || !Array.isArray(json)) {
                        throw new Error("Api error: GET " + path + " should return an object array.");
                    }
                    if (json.length) {
                        return createRecord(json[0]);
                    }
                    return null;
                };
                break;
            case 'count':
                reducer = function (json) {
                    if (!json || typeof json !== 'object' || !json.hasOwnProperty('count')) {
                        throw new Error("Api error: GET " + path + " should return an object with count property.");
                    }
                    return json.count;
                };
                break;
            case 'findByPk':
            case 'create':
            case 'watch':
                reducer = function (json) { return (json ? createRecord(json) : null); };
                break;
            case 'find':
                reducer = function (json) {
                    if (!Array.isArray(json))
                        throw new Error("Api error: GET " + path + " should return an object array.");
                    return json.map(createRecord);
                };
                break;
            case 'paginate':
                reducer = function (json) {
                    if (json === null || json === void 0 ? void 0 : json.results) {
                        json.results = json.results.map(createRecord);
                    }
                    return json;
                };
                break;
            default:
        }
        this._result = M.request(path, init, this, reducer);
        return this._result;
    };
    Query.prototype._debug = function () {
        if (debug.enabled) {
            var str = '';
            var M = this.model;
            if (M.name === 'AnonymousModel') {
                str = "Client(\"" + M.path + "\").";
            }
            else {
                str = M.name + ".";
            }
            str += this._op;
            switch (this._op) {
                case 'findByPk':
                    if (this._id === null) {
                        throw new Error('id is not specified for findByPk');
                    }
                    str += "(" + JSON.stringify(this._id) + ")";
                    break;
                case 'remove':
                    if (this._id !== null) {
                        str += "(" + JSON.stringify(this._id) + ")";
                    }
                    else {
                        str += '()';
                    }
                    break;
                case 'update':
                    if (this._id) {
                        str += "(" + JSON.stringify(this._id) + ", " + JSON.stringify(this._data) + ")";
                    }
                    else {
                        str += "(" + JSON.stringify(this._data) + ")";
                    }
                    break;
                case 'create':
                    str += "(" + JSON.stringify(this._data) + ")";
                    break;
                default:
                    str += '()';
            }
            if (this._args) {
                var args_1 = this._args;
                str += Object.keys(this._args)
                    .map(function (key) { return ".arg(\"" + key + "\", " + JSON.stringify(args_1[key]) + ")"; })
                    .join('');
            }
            if (this._filters) {
                str += ".where(" + JSON.stringify(this._filters) + ")";
            }
            if (this._search) {
                str += ".search(\"" + this._search + "\")";
            }
            if (this._limit > 1) {
                str += ".limit(" + this._limit + ")";
            }
            if (this._page > 1) {
                str += ".page(" + this._page + ")";
            }
            if (this._sort) {
                str += ".sort(" + this._sort + ")";
            }
            debug(str);
        }
    };
    Query.prototype._createInit = function () {
        var init = {};
        init.method = 'GET';
        var path = '';
        var query = {};
        if (this._args) {
            var obj_1 = this._args;
            Object.keys(obj_1).forEach(function (key) {
                query["_" + key] = obj_1[key];
            });
        }
        if (this._filters) {
            Object.assign(query, this._filters);
        }
        if (this._search) {
            query._search = this._search;
        }
        if (this._limit) {
            query._limit = this._limit;
        }
        if (this._page) {
            query._page = this._page;
        }
        if (this._sort) {
            query._sort = this._sort;
        }
        if (this._id !== null && ['findByPk', 'remove', 'update'].indexOf(this._op) > -1) {
            path += "/" + encodeURIComponent(this._id);
        }
        if (this._data && ['create', 'update'].indexOf(this._op) > -1) {
            init.body = this._data;
        }
        switch (this._op) {
            case 'count':
                path += '/count';
                break;
            case 'paginate':
                path += '/paginate';
                break;
            case 'watch':
                path += '/watch';
                break;
            case 'create':
                init.method = 'POST';
                break;
            case 'update':
                init.method = 'PATCH';
                break;
            case 'remove':
                init.method = 'DELETE';
                break;
            default:
        }
        init.path = path;
        init.query = query;
        return init;
    };
    Query.prototype.then = function (onSuccess, onFail) {
        return this.exec().then(onSuccess, onFail);
    };
    Query.prototype.catch = function (onFail) {
        return this.exec().catch(onFail);
    };
    Query.prototype.finally = function (fn) {
        return this.exec().finally(fn);
    };
    return Query;
}());
exports.default = Query;
